# -*- coding: utf-8 -*-

import unittest
import boto

TABLE_NAME = 'Table-HR'
TABLE_NAME_404 = 'Waldo'
TABLE_RT = 45
TABLE_WT = 123
TABLE_RT2 = 10
TABLE_WT2 = 10
TABLE_HK_NAME = u'hash_key'
TABLE_HK_TYPE = u'N'
TABLE_RK_NAME = u'range_key'
TABLE_RK_TYPE = u'S'

HK_VALUE1 = u'123'
HK_VALUE2 = u'456'
HK_VALUE3 = u'789'
RK_VALUE1 = u'Waldo-1'
RK_VALUE2 = u'Waldo-2'
RK_VALUE3 = u'Waldo-3'
RK_VALUE4 = u'Waldo-4'
RK_VALUE5 = u'Waldo-5'
RK_VALUE6 = u'Waldo-6'


ITEM1 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE1},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE1},
    u'relevant_data': {u'S': u'tata'},
}
ITEM2 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE1},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE2},
    u'relevant_data': {u'S': u'tete'},
}
ITEM3 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE2},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE3},
    u'relevant_data': {u'S': u'titi'},
}
ITEM4 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE3},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE4},
    u'relevant_data': {u'S': u'toto'},
}
ITEM5 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE3},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE5},
    u'relevant_data': {u'S': u'tutu'},
}
ITEM6 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE3},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE6},
    u'relevant_data': {u'S': u'tyty'},
    u'irelevant_data': {u'S': u'a'*(2048)},
}

# Please note that most query features are not yet implemented hence not tested
class TestScan(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.db import dynamodb
        from ddbmock.database.table import Table
        from ddbmock.database.key import PrimaryKey

        dynamodb.hard_reset()

        hash_key = PrimaryKey(TABLE_HK_NAME, TABLE_HK_TYPE)
        range_key = PrimaryKey(TABLE_RK_NAME, TABLE_RK_TYPE)

        self.t1 = Table(TABLE_NAME, TABLE_RT, TABLE_WT, hash_key, range_key)

        dynamodb.data[TABLE_NAME]  = self.t1

        self.t1.put(ITEM1, {})
        self.t1.put(ITEM2, {})
        self.t1.put(ITEM3, {})
        self.t1.put(ITEM4, {})
        self.t1.put(ITEM5, {})
        self.t1.put(ITEM6, {})

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        from ddbmock import clean_boto_patch
        dynamodb.hard_reset()
        clean_boto_patch()

    def test_scan_all(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        expected = {
            u"Count": 6,
            u"ScannedCount": 6,
            u"Items": [ITEM2, ITEM1, ITEM6, ITEM5, ITEM4, ITEM3],
            u"ConsumedCapacityUnits": 1.5,
        }

        db = connect_boto_patch()

        ret = db.layer1.scan(TABLE_NAME, None)
        self.assertEqual(expected, ret)

    def test_scan_paged(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.dynamodb.exceptions import DynamoDBValidationError

        esk = {
            u'HashKeyElement': {u'N': u'789'},
            u'RangeKeyElement': {u'S': u'Waldo-5'},
        }

        expected1 = {
            u"Count": 4,
            u"ScannedCount": 4,
            u"Items": [ITEM2, ITEM1, ITEM6, ITEM5],
            u"ConsumedCapacityUnits": 1.5,
            u'LastEvaluatedKey': esk,
        }
        expected2 = {
            u"Count": 2,
            u"ScannedCount": 2,
            u"Items": [ITEM4, ITEM3],
            u"ConsumedCapacityUnits": 0.5,
        }

        db = connect_boto_patch()

        ret = db.layer1.scan(TABLE_NAME, limit=4)
        self.assertEqual(expected1, ret)
        ret = db.layer1.scan(TABLE_NAME, exclusive_start_key=esk)
        self.assertEqual(expected2, ret)

    def test_scan_all_filter_fields(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        self.maxDiff = None

        expected = {
            u"Count": 6,
            u"ScannedCount": 6,
            u"Items": [
                {u"relevant_data": {u"S": u"tete"}},
                {u"relevant_data": {u"S": u"tata"}},
                {u"relevant_data": {u"S": u"tyty"}},
                {u"relevant_data": {u"S": u"tutu"}},
                {u"relevant_data": {u"S": u"toto"}},
                {u"relevant_data": {u"S": u"titi"}},
            ],
            u"ConsumedCapacityUnits": 1.5,
        }
        fields = [u'relevant_data']

        db = connect_boto_patch()

        ret = db.layer1.scan(TABLE_NAME, None, fields)
        self.assertEqual(expected, ret)

    # No need to test all conditions/type mismatch as they are unit tested
    def test_scan_condition_filter_fields_in(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        expected = {
            u"Count": 3,
            u"ScannedCount": 6,
            u"Items": [
                {u"relevant_data": {u"S": u"tata"}},
                {u"relevant_data": {u"S": u"toto"}},
                {u"relevant_data": {u"S": u"titi"}},
            ],
            u"ConsumedCapacityUnits": 1.5,
        }

        conditions = {
            "relevant_data": {
                "AttributeValueList": [{"S":"toto"},{"S":"titi"},{"S":"tata"}],
                "ComparisonOperator": "IN",
            }
        }
        fields = [u'relevant_data']

        db = connect_boto_patch()

        ret = db.layer1.scan(TABLE_NAME, conditions, fields)
        self.assertEqual(expected, ret)

    def test_scan_condition_filter_fields_contains(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        expected = {
            u"Count": 1,
            u"ScannedCount": 6,
            u"Items": [
                {u"relevant_data": {u"S": u"toto"}},
            ],
            u"ConsumedCapacityUnits": 1.5,
        }

        conditions = {
            "relevant_data": {
                "AttributeValueList": [{"S":"to"}],
                "ComparisonOperator": "CONTAINS",
            },
        }
        fields = [u'relevant_data']

        db = connect_boto_patch()

        ret = db.layer1.scan(TABLE_NAME, conditions, fields)
        self.assertEqual(expected, ret)

    def test_scan_filter_ghost_fields(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        expected = {
            u"Count": 0,
            u"ScannedCount": 6,
            u"Items": [],
            u"ConsumedCapacityUnits": 1.5,
        }

        conditions = {
            "ghost field": {
                "AttributeValueList": [{"N":"123"}],
                "ComparisonOperator": "LT",
            },
        }
        fields = [u'relevant_data']

        db = connect_boto_patch()

        ret = db.layer1.scan(TABLE_NAME, conditions, fields)
        self.assertEqual(expected, ret)

    def test_scan_validation_error(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.dynamodb.exceptions import DynamoDBValidationError

        expected = {
            u"Count": 1,
            u"ScannedCount": 6,
            u"Items": [
                {u"relevant_data": {u"S": u"toto"}},
            ],
            u"ConsumedCapacityUnits": 1.5,
        }

        conditions = {
            "relevant_data": {
                "AttributeValueList": [{"S":"to"},{"S":"ta"}],
                "ComparisonOperator": "CONTAINS",
            }
        }
        fields = [u'relevant_data']

        db = connect_boto_patch()

        self.assertRaises(DynamoDBValidationError, db.layer1.scan,
            TABLE_NAME, conditions, fields
        )



# -*- coding: utf-8 -*-

import unittest, mock, json, time

NOW = time.time()

TABLE_NAME1 = 'Table-1'

TABLE_RT = 45
TABLE_WT = 123

HASH_KEY = {"AttributeName":"hash_key","AttributeType":"N"}
RANGE_KEY = {"AttributeName":"range_key","AttributeType":"S"}

HEADERS = {
    'x-amz-target': 'dynamodb_20111205.CreateTable',
    'content-type': 'application/x-amz-json-1.0',
}

# Goal here is not to test the full API, this is done by the Boto tests
class TestCreateTable(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.db import dynamodb
        from ddbmock import main
        app = main({})
        from webtest import TestApp
        self.app = TestApp(app)
        dynamodb.hard_reset()

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        dynamodb.hard_reset()

    @mock.patch("ddbmock.database.table.time")
    def test_create_table_hr(self, m_time):
        from ddbmock.database.db import dynamodb

        m_time.time.return_value = NOW

        request = {
            "TableName": TABLE_NAME1,
            "KeySchema": {
                "HashKeyElement": HASH_KEY,
                "RangeKeyElement": RANGE_KEY,
            },
            "ProvisionedThroughput": {
                "ReadCapacityUnits": TABLE_RT,
                "WriteCapacityUnits": TABLE_WT,
            }
        }

        expected = {
            u'TableDescription': {
                u'CreationDateTime': NOW,
                u'KeySchema': {
                    u'HashKeyElement': HASH_KEY,
                    u'RangeKeyElement': RANGE_KEY,
                },
                u'ProvisionedThroughput': {
                    u'ReadCapacityUnits': TABLE_RT,
                    u'WriteCapacityUnits': TABLE_WT,
                },
                u'TableName': TABLE_NAME1,
                u'TableStatus': u'CREATING',
            }
        }

        res = self.app.post_json('/', request, HEADERS, status=200)
        self.assertEqual(expected, json.loads(res.body))
        self.assertEqual('application/x-amz-json-1.0; charset=UTF-8', res.headers['Content-Type'])

        data = dynamodb.data
        assert TABLE_NAME1 in data
        table = data[TABLE_NAME1]

        self.assertEqual(TABLE_NAME1, table.name)
        self.assertEqual(TABLE_RT, table.rt)
        self.assertEqual(TABLE_WT, table.wt)
        self.assertEqual(NOW, table.creation_time)
        self.assertEqual(HASH_KEY['AttributeName'], table.hash_key.name)
        self.assertEqual(RANGE_KEY['AttributeName'], table.range_key.name)
        self.assertEqual(HASH_KEY['AttributeType'], table.hash_key.typename)
        self.assertEqual(RANGE_KEY['AttributeType'], table.range_key.typename)

    # The real goal of this test is to validate the error view. The tested behavior
    # is already known to work thanks the boto tests
    def test_create_table_twice_fails(self):
        from ddbmock.database.db import dynamodb

        request = {
            "TableName": TABLE_NAME1,
            "KeySchema": {
                "HashKeyElement": HASH_KEY,
                "RangeKeyElement": RANGE_KEY,
            },
            "ProvisionedThroughput": {
                "ReadCapacityUnits": TABLE_RT,
                "WriteCapacityUnits": TABLE_WT,
            }
        }

        expected = {
            u'__type': u'com.amazonaws.dynamodb.v20111205#ResourceInUseException',
            u'message': u'Table {} already exists'.format(TABLE_NAME1),
        }


        res = self.app.post_json('/', request, HEADERS, status=200)
        res = self.app.post_json('/', request, HEADERS, status=400)
        self.assertEqual('application/x-amz-json-1.0; charset=UTF-8', res.headers['Content-Type'])
        self.assertEqual(expected, json.loads(res.body))



# -*- coding: utf-8 -*-

import unittest
import boto

TABLE_NAME = 'Table-HR'
TABLE_NAME2 = 'Table-H'
TABLE_NAME_404 = 'Waldo'
TABLE_RT = 45
TABLE_WT = 123
TABLE_RT2 = 10
TABLE_WT2 = 10
TABLE_HK_NAME = u'hash_key'
TABLE_HK_TYPE = u'N'
TABLE_RK_NAME = u'range_key'
TABLE_RK_TYPE = u'S'

HK_VALUE = u'123'
HK_VALUE2 = u'456'
HK_VALUE_404 = u'404'
RK_VALUE = u'Decode this data if you are a coder'


ITEM = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE},
    u'relevant_data': {u'B': u'THVkaWEgaXMgdGhlIGJlc3QgY29tcGFueSBldmVyIQ=='},
}
ITEM2 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    u'relevant_data': {u'B': u'THVkaWEgaXMgdGhlIGJlc3QgY29tcGFueSBldmVyIQ=='},
}
ITEM_BIG = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE2},
    u'relevant_data': {u'S': u'a'*1024},
}

class TestGetItem(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.db import dynamodb
        from ddbmock.database.table import Table
        from ddbmock.database.key import PrimaryKey

        dynamodb.hard_reset()

        hash_key = PrimaryKey(TABLE_HK_NAME, TABLE_HK_TYPE)
        range_key = PrimaryKey(TABLE_RK_NAME, TABLE_RK_TYPE)

        self.t1 = Table(TABLE_NAME, TABLE_RT, TABLE_WT, hash_key, range_key)
        self.t2 = Table(TABLE_NAME2, TABLE_RT, TABLE_WT, hash_key, None)

        dynamodb.data[TABLE_NAME]  = self.t1
        dynamodb.data[TABLE_NAME2] = self.t2

        self.t1.put(ITEM, {})
        self.t2.put(ITEM2, {})
        self.t2.put(ITEM_BIG, {})

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        from ddbmock import clean_boto_patch
        dynamodb.hard_reset()
        clean_boto_patch()

    def test_get_hr(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        db = connect_boto_patch()

        expected = {
            u'ConsumedCapacityUnits': 0.5,
            u'Item': ITEM,
        }

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
            u"RangeKeyElement": {TABLE_RK_TYPE: RK_VALUE},
        }

        self.assertEquals(expected, db.layer1.get_item(TABLE_NAME, key))

    def test_get_hr_consistent(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        db = connect_boto_patch()

        expected = {
            u'ConsumedCapacityUnits': 1,
            u'Item': ITEM,
        }

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
            u"RangeKeyElement": {TABLE_RK_TYPE: RK_VALUE},
        }

        self.assertEquals(expected, db.layer1.get_item(TABLE_NAME, key, consistent_read=True))

    def test_get_h(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        db = connect_boto_patch()

        expected = {
            u'ConsumedCapacityUnits': 0.5,
            u'Item': ITEM2,
        }

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
        }

        self.assertEquals(expected, db.layer1.get_item(TABLE_NAME2, key))

    def test_get_consistent_big_attributes_to_get(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        db = connect_boto_patch()

        expected = {
            u'ConsumedCapacityUnits': 2,
            u'Item': {TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE2}},
        }

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE2},
        }

        self.assertEquals(expected, db.layer1.get_item(TABLE_NAME2, key, consistent_read=True, attributes_to_get=[TABLE_HK_NAME]))

    def test_get_h_404(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.dynamodb.exceptions import DynamoDBKeyNotFoundError

        db = connect_boto_patch()

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE_404},
        }

        self.assertRaises(DynamoDBKeyNotFoundError,
                          db.layer1.get_item,
                          TABLE_NAME2, key)

    def test_get_hr_attr_to_get(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        db = connect_boto_patch()

        expected = {
            u'ConsumedCapacityUnits': 0.5,
            u'Item': {
                u'relevant_data': {u'B': u'THVkaWEgaXMgdGhlIGJlc3QgY29tcGFueSBldmVyIQ=='}}
        }

        key = {
            u"HashKeyElement":  {TABLE_HK_TYPE: HK_VALUE},
            u"RangeKeyElement": {TABLE_RK_TYPE: RK_VALUE},
        }

        self.assertEquals(expected, db.layer1.get_item(TABLE_NAME, key, attributes_to_get=[u'relevant_data']))



# -*- coding: utf-8 -*-

from .types import table_name, Required, get_key_schema, attributes_to_get_schema,consistent_read

post = {
    u"RequestItems": {
        table_name: {
            u"Keys": [get_key_schema],
            Required(u'AttributesToGet', []): attributes_to_get_schema,
            Required(u'ConsistentRead', False): consistent_read,
        },
    },
}



# -*- coding: utf-8 -*-

import unittest, mock

# tests
# - update throughput

NAME = "tabloid"
RT = 100
WT = 100
RT2 = 120
WT2 = 120
RT3 = 20
WT3 = 20
RT4 = 2
WT4 = 2
RT10 = 109
WT10 = 109
RT100 = 201
WT100 = 201
CREATION = 2*24*3600  # day 2 of our era (UNIX) (avoids stupid side effects of 0 in the test)

class TestTable(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.table import Table
        self.table = Table(NAME, RT, WT, None, None, status="ACTIVE")

    def tearDown(self):
        self.table = None

    def test_update_throughput_nominal(self):
        self.table.update_throughput(RT2, WT2)

        self.assertEqual(RT2, self.table.rt)
        self.assertEqual(WT2, self.table.wt)

    def test_increase_throughtput_max_100_percents(self):
        from ddbmock.errors import LimitExceededException

        # Try on RT
        self.assertRaisesRegexp(LimitExceededException,
                                "ReadCapacityUnits .* at most 100",
                                self.table.update_throughput,
                                RT100, WT)
        self.assertEqual(RT, self.table.rt)
        self.assertEqual(WT, self.table.wt)

        # Try on WT
        self.assertRaisesRegexp(LimitExceededException,
                                "WriteCapacityUnits .* at most 100",
                                self.table.update_throughput,
                                RT, WT100)
        self.assertEqual(RT, self.table.rt)
        self.assertEqual(WT, self.table.wt)

    @mock.patch("ddbmock.database.table.time")
    def test_decrease_max_once_a_day(self, m_time):
        from ddbmock.errors import LimitExceededException

        # 1st decrease 1 hour after creation (ok)
        m_time.time.return_value = CREATION + 1*3600
        self.table.update_throughput(RT3, WT3)
        self.assertEqual(RT3, self.table.rt)
        self.assertEqual(WT3, self.table.wt)

        # bypass the timer
        self.table.status = "ACTIVE"

        # 2nd decrease 2 hour after creation (fail)
        m_time.time.return_value = CREATION + 2*3600
        self.assertRaisesRegexp(LimitExceededException,
                                "once .* day",
                                self.table.update_throughput,
                                RT4, WT4)
        self.assertEqual(RT3, self.table.rt)
        self.assertEqual(WT3, self.table.wt)







# -*- coding: utf-8 -*-

from ddbmock.database import dynamodb

def delete_table(post):
    name = post[u'TableName']
    table = dynamodb.delete_table(name)

    return {
        'TableDescription': table.to_dict(verbose=False)
    }



# -*- coding: utf-8 -*-

import unittest

# tests
# - actions
# - expected values
# - size computation (fields)
# - size computation (item)

FIELDNAME = "fieldname"
VALUE_S = {u"S": u"String value ê"}
VALUE_N = {u"N": u"123"}
VALUE_B = {u"B": u"QmluYXJ5IGRhdGE=="}
VALUE_SS = {u"SS": [u"String value", u"Ludia", u"Waldo ç"]}
VALUE_NS = {u"NS": [u"456", u"789"]}
VALUE_BS = {u"BS": [u"QmluYXJ5IGRhdGE==", u"enV0IGRlIHp1dA=="]}

VALUE_SS_DEL = {u"SS": [u"Ludia", u"zinga"]}
VALUE_SS_SUR = {u"SS": [u"String value", u"Waldo ç"]}
VALUE_SS_RES = {u"SS": [u'Ludia', u'Waldo ç', u'String value', u'zinga']}
VALUE_N_X2 = {u"N": u"246"}

VALUE_S_BIG = {u"S": u"abcd"*31415}  # if you don't see the geek ref here, your not a geek

ITEM_TYPE = {
    u'S': VALUE_S,
    u'B': VALUE_B,
    u'N': VALUE_N,
    u'SS': VALUE_SS,
    u'BS': VALUE_BS,
    u'NS': VALUE_NS,
}

ITEM_BIG = {
    u'big field': VALUE_S_BIG,
}

class TestItem(unittest.TestCase):
    # actions
    def test_action_put(self):
        from ddbmock.database.item import Item

        item = Item({})
        item._apply_action(FIELDNAME, {"Action": "PUT", "Value": VALUE_S})
        self.assertEqual([(FIELDNAME, VALUE_S)], item.items())

    def test_action_delete(self):
        from ddbmock.database.item import Item

        # field does not exist (just ignores)
        item = Item({})
        item._apply_action(FIELDNAME, {"Action": "DELETE", "Value": VALUE_S})
        self.assertEqual([], item.items())

        # delete scalar value
        item = Item({
            FIELDNAME: VALUE_S,
        })
        item._apply_action(FIELDNAME, {"Action": "DELETE"})
        self.assertEqual([], item.items())

        # delete set from set
        item = Item({
            FIELDNAME: VALUE_SS,
        })
        item._apply_action(FIELDNAME, {"Action": "DELETE", "Value": VALUE_SS_DEL})
        self.assertEqual([(FIELDNAME, VALUE_SS_SUR)], item.items())

        # delete set all from set
        item = Item({
            FIELDNAME: VALUE_SS,
        })
        item._apply_action(FIELDNAME, {"Action": "DELETE", "Value": VALUE_SS})
        self.assertEqual([], item.items())

    def test_action_add(self):
        from ddbmock.database.item import Item

        # add value to int field
        item = Item({FIELDNAME: VALUE_N})
        item._apply_action(FIELDNAME, {"Action": "ADD", "Value": VALUE_N})
        self.assertEqual([(FIELDNAME, VALUE_N_X2)], item.items())

        # push value to any set
        item = Item({FIELDNAME: VALUE_SS})
        item._apply_action(FIELDNAME, {"Action": "ADD", "Value": VALUE_SS_DEL})
        self.assertEqual([(FIELDNAME, VALUE_SS_RES)], item.items())

        # field does not exist
        item = Item({})
        item._apply_action(FIELDNAME, {"Action": "ADD", "Value": VALUE_NS})
        self.assertEqual([(FIELDNAME, VALUE_NS)], item.items())


    def test_action_type_mismatch(self):
        from ddbmock.database.item import Item
        from ddbmock.errors import ValidationException

        # delete from set, set of same type
        item = Item({
            FIELDNAME: VALUE_SS,
        })
        self.assertRaises(ValidationException,
                          item._apply_action,
                          FIELDNAME, {"Action": "DELETE", "Value": VALUE_S})
        self.assertEqual([(FIELDNAME, VALUE_SS)], item.items())
        self.assertRaises(ValidationException,
                          item._apply_action,
                          FIELDNAME, {"Action": "DELETE", "Value": VALUE_NS})
        self.assertEqual([(FIELDNAME, VALUE_SS)], item.items())

        # delete value from scalar field (fail)
        item = Item({
            FIELDNAME: VALUE_S,
        })
        self.assertRaises(ValidationException,
                          item._apply_action,
                          FIELDNAME, {"Action": "DELETE", "Value": VALUE_S})
        self.assertEqual([(FIELDNAME, VALUE_S)], item.items())

        # add to scalar non number field
        item = Item({FIELDNAME: VALUE_S})
        self.assertRaises(ValidationException,
                          item._apply_action,
                          FIELDNAME, {"Action": "ADD", "Value": VALUE_S})
        self.assertEqual([(FIELDNAME, VALUE_S)], item.items())

        # add to set of different type
        item = Item({FIELDNAME: VALUE_SS})
        self.assertRaises(ValidationException,
                          item._apply_action,
                          FIELDNAME, {"Action": "ADD", "Value": VALUE_NS})
        self.assertEqual([(FIELDNAME, VALUE_SS)], item.items())

        # add non number to non existing field
        item = Item({})
        self.assertRaises(ValidationException,
                          item._apply_action,
                          FIELDNAME, {"Action": "ADD", "Value": VALUE_S})
        self.assertEqual([], item.items())

    def test_expected_field_does_not_exist(self):
        from ddbmock.database.item import Item
        from ddbmock.errors import ConditionalCheckFailedException

        # ok
        item = Item({})
        item.assert_match_expected({
           FIELDNAME: {u'Exists': False}
        })

        # mismatch
        item = Item({FIELDNAME: VALUE_SS})
        self.assertRaises(ConditionalCheckFailedException,
                          item.assert_match_expected,
                          {FIELDNAME: {u'Exists': False}})

    def test_expected_field_exists_value(self):
        from ddbmock.database.item import Item
        from ddbmock.errors import ConditionalCheckFailedException

        # ok
        item = Item({FIELDNAME: VALUE_S})
        item.assert_match_expected({
           FIELDNAME: {'Exists': True, 'Value': VALUE_S}
        })

        # exitence mismatch
        item = Item({})
        self.assertRaises(ConditionalCheckFailedException,
                          item.assert_match_expected,
                          {FIELDNAME: {'Exists': True, 'Value': VALUE_S}})

        # type mismatch
        item = Item({FIELDNAME: VALUE_S})
        self.assertRaises(ConditionalCheckFailedException,
                          item.assert_match_expected,
                          {FIELDNAME: {'Exists': True, 'Value': VALUE_N}})

    def test_size_computation(self):
        from ddbmock.database.item import Item

        item = Item(ITEM_TYPE)

        self.assertEqual(0, item.get_field_size(u"toto"))
        self.assertEqual(8, item.get_field_size(u"N"))
        self.assertEqual(15, item.get_field_size(u"S"))
        self.assertEqual(12, item.get_field_size(u"B"))
        self.assertEqual(16, item.get_field_size(u"NS"))
        self.assertEqual(30, item.get_field_size(u"SS"))
        self.assertEqual(24, item.get_field_size(u"BS"))

    def test_item_size_computation(self):
        from ddbmock.database.item import Item

        item1 = Item()
        item2 = Item(ITEM_TYPE)

        # cache init
        self.assertIsNone(item1.size)
        self.assertIsNone(item2.size)

        # compute size
        s1 = item1.get_size()
        s2 = item2.get_size()

        self.assertEqual(0, s1)
        self.assertEqual(114, s2)

        # check cache
        self.assertEqual(s1, item1.size)
        self.assertEqual(s2, item2.size)

        # any call to "update item" invalidates the cache
        item1.apply_actions({})
        item2.apply_actions({})

        self.assertIsNone(item1.size)
        self.assertIsNone(item2.size)

class TestItemSize(unittest.TestCase):
    def test_unit_from_size(self):
        from ddbmock.database.item import ItemSize

        self.assertEqual(1, ItemSize(0).as_units())  # not less than 1 unit
        self.assertEqual(1, ItemSize(1).as_units())
        self.assertEqual(1, ItemSize(1024).as_units())
        self.assertEqual(2, ItemSize(1025).as_units())
        self.assertEqual(3, ItemSize(2049).as_units())

    def test_unit_add(self):
        from ddbmock.database.item import ItemSize

        self.assertEqual(ItemSize(4), ItemSize(1)+ItemSize(3))
        self.assertEqual(ItemSize, type(ItemSize(1)+ItemSize(3)))

    def test_size_with_index(self):
        from ddbmock.database.item import ItemSize

        self.assertEqual(0+100, ItemSize(0).with_indexing_overhead())
        self.assertEqual(1+100, ItemSize(1).with_indexing_overhead())
        self.assertEqual(1024+100, ItemSize(1024).with_indexing_overhead())
        self.assertEqual(2049+100, ItemSize(2049).with_indexing_overhead())

    def test_item_unit_computation(self):
        # full chain test (~functional)
        from ddbmock.database.item import Item

        item1 = Item(ITEM_BIG)
        item2 = Item(ITEM_TYPE)

        self.assertEqual(123, item1.get_size().as_units())  # If I tell you the '123' is not on purpose, you won't believe me, will you ? Especially when Pi is involved
        self.assertEqual(1, item2.get_size().as_units())




# -*- coding: utf-8 -*-

import unittest, mock

NAME = "TableName"
HASH = "hash_key"
HASH_404 = "hash_key 404"
RANGE1 = "range_key 1"
RANGE2 = "range_key 2"
RANGE_404 = "range_key 404"
DATA1 = {"key": "value 1"}
DATA2 = {"key": "value 2"}

class TestMemoryStore(unittest.TestCase):
    def test_truncate(self):
        from ddbmock.database.storage.memory import Store

        ms = Store(NAME)

        ms.data[1] = "some data"
        ms.data[2] = "some data"
        ms.data[3] = "some data"
        ms.data[4] = "some data"
        ms.data[5] = "some data"

        ms.truncate()

        self.assertFalse(ms.data)

    def test_set_item(self):
        from ddbmock.database.storage.memory import Store

        ms = Store(NAME)

        ms[HASH,RANGE1] = DATA1
        self.assertEqual(DATA1, ms.data[HASH][RANGE1])

    def test_get_item(self):
        from ddbmock.database.storage.memory import Store

        ms = Store(NAME)

        ms.data[HASH][RANGE1] = DATA1
        ms.data[HASH][RANGE2] = DATA2

        self.assertRaises(KeyError, ms.__getitem__, (HASH_404, None))
        self.assertRaises(KeyError, ms.__getitem__, (HASH_404, RANGE1))
        self.assertRaises(KeyError, ms.__getitem__, (HASH_404, RANGE1))
        self.assertRaises(KeyError, ms.__getitem__, (HASH, RANGE_404))

        self.assertEqual(DATA1, ms[HASH, RANGE1])
        self.assertEqual(DATA2, ms[HASH, RANGE2])
        self.assertEqual({RANGE1:DATA1, RANGE2:DATA2}, ms[HASH, None])

    def test_del_item(self):
        from ddbmock.database.storage.memory import Store

        ms = Store(NAME)

        ms.data[HASH][RANGE1] = DATA1
        ms.data[HASH][RANGE2] = DATA2

        self.assertRaises(KeyError, ms.__delitem__, (HASH_404, None))
        self.assertRaises(KeyError, ms.__delitem__, (HASH_404, RANGE1))
        self.assertRaises(KeyError, ms.__delitem__, (HASH_404, RANGE1))
        self.assertRaises(KeyError, ms.__delitem__, (HASH, RANGE_404))

        del ms[HASH, RANGE1]
        del ms[HASH, RANGE2]
        self.assertNotIn(RANGE1, ms.data[HASH])
        self.assertNotIn(RANGE2, ms.data[HASH])

    def test_iter(self):
        from ddbmock.database.storage.memory import Store

        ms = Store(NAME)

        ms.data[HASH][RANGE1] = DATA1
        ms.data[HASH][RANGE2] = DATA2

        self.assertEqual([DATA1, DATA2], list(ms))



# -*- coding: utf-8 -*-

from ddbmock.database import dynamodb

def create_table(post):
    table = dynamodb.create_table(post[u'TableName'], post)

    return {
        "TableDescription": table.to_dict(verbose=False),
    }



# -*- coding: utf-8 -*-

# hack for boto integration tests to run
# the test module expects to find `unitest` loaded as part of tests.unit
# and, as we have the same name (collisions happiness), I cam not use regular
# and easy import. :')

import unittest



# -*- coding: utf-8 -*-

from .types import table_name, table_key_schema, provisioned_throughtput

post = {
    u'TableName': table_name,
    u'KeySchema': table_key_schema,
    u'ProvisionedThroughput': provisioned_throughtput,
}



# -*- coding: utf-8 -*-

import unittest
import boto

TABLE_NAME1 = 'Table-1'
TABLE_NAME2 = 'Table-2'

class TestListTables(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.db import dynamodb
        from ddbmock.database.table import Table
        from ddbmock.database.key import PrimaryKey

        dynamodb.hard_reset()

        hash_key = PrimaryKey('hash_key', 'N')
        range_key = PrimaryKey('range_key', 'S')

        t1 = Table(TABLE_NAME1, 10, 10, hash_key, range_key)
        t2 = Table(TABLE_NAME2, 10, 10, hash_key, range_key)

        dynamodb.data[TABLE_NAME1] = t1
        dynamodb.data[TABLE_NAME2] = t2

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        from ddbmock import clean_boto_patch
        dynamodb.hard_reset()
        clean_boto_patch()

    def test_list_tables(self):
        from ddbmock import connect_boto_patch
        db = connect_boto_patch()

        expected = [TABLE_NAME1, TABLE_NAME2]

        self.assertEqual(expected, db.list_tables())




# -*- coding: utf-8 -*-

from .types import table_name, Required, item_schema, return_values, expected_schema

post = {
    u'TableName': table_name,
    u'Item': item_schema,
    Required(u'Expected', {}): expected_schema, # It is optional but with a def value
    Required(u'ReturnValues', u'NONE'): return_values, # It is optional but with a def value
}



# -*- coding: utf-8 -*-

from . import load_table
from ddbmock.utils import push_write_throughput
from ddbmock.errors import ValidationException

@load_table
def scan(post, table):
    if post[u'AttributesToGet'] and post[u'Count']:
        raise ValidationException("Can not filter fields when only count is requested")

    results = table.scan(
        post[u'ScanFilter'],
        post[u'AttributesToGet'],
        post[u'ExclusiveStartKey'],
        post[u'Limit'],
    )

    capacity = 0.5*results.size.as_units()
    push_write_throughput(table.name, capacity)

    ret = {
        "Count": len(results.items),
        "ScannedCount": results.scanned,
        "ConsumedCapacityUnits": capacity,
    }

    if results.last_key:
        ret['LastEvaluatedKey'] = results.last_key

    if not post[u'Count']:
        ret[u'Items'] = results.items

    return ret



# -*- coding: utf-8 -*-

import unittest, mock

ACTION = "CreateTable"
ACTION_404 = "!~I bet this route won't ever exist~!"
ROUTE = "create_table"
POST = {"toto":"titi"}

class TestRouterInit(unittest.TestCase):
    @mock.patch("ddbmock.router.import_module")
    @mock.patch("ddbmock.router.dynamodb_api_validate")
    def test_do_request_nominal(self, m_validate, m_import):
        from ddbmock.router import router

        m_module = m_import.return_value
        m_route = m_module.create_table
        m_route.return_value = {}
        m_validated = m_validate.return_value

        router(ACTION, POST)

        m_import.assert_called_with("ddbmock.operations.{}".format(ROUTE))
        m_validate.assert_called_with(ROUTE, POST)
        m_route.assert_called_with(m_validated)

    def test_do_request_route_404(self):
        from ddbmock.router import router
        from ddbmock.errors import InternalFailure

        self.assertRaisesRegexp(InternalFailure,
                                'exist',
                                router,
                                ACTION_404, POST)

    @mock.patch("ddbmock.router.import_module")
    @mock.patch("ddbmock.router.dynamodb_api_validate")
    def test_internal_server_error(self, m_validate, m_import):
        from ddbmock.router import router
        from ddbmock.errors import InternalFailure

        m_module = m_import.return_value
        m_route = m_module.create_table
        m_route.return_value = {}
        m_route.side_effect = ValueError
        m_validated = m_validate.return_value

        self.assertRaisesRegexp(InternalFailure,
                                'ValueError',
                                router,
                                ACTION, POST)

        m_import.assert_called_with("ddbmock.operations.{}".format(ROUTE))
        m_validate.assert_called_with(ROUTE, POST)
        m_route.assert_called_with(m_validated)



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup_requires = [
    'd2to1',
    'boto',
    'nose',
    'nosexcover',
    'coverage',
    'mock',
    'webtest',
    'Sphinx',
]

setup(
    d2to1=True,
    keywords='pyramid dynamodb mock planification',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    setup_requires=setup_requires,
    test_suite='nose.collector',
    entry_points="""\
    [paste.app_factory]
    main = ddbmock:main
    """,
)



# -*- coding: utf-8 -*-

from ddbmock.database import dynamodb

def batch_get_item(post):
    #TODO: limit to 100/batch
    #TODO: unprocessed keys

    return {
        "Responses": dynamodb.get_batch(post[u'RequestItems']),
    }




# -*- coding: utf-8 -*-

import unittest, mock

class TestRouterBoto(unittest.TestCase):
    def test_routing_exception_translation(self):
        from ddbmock.router.boto import _ddbmock_exception_to_boto_exception
        from boto.exception import DynamoDBResponseError
        from ddbmock.errors import InternalFailure

        self.assertIsInstance(_ddbmock_exception_to_boto_exception(
                              InternalFailure('taratata')),
                              DynamoDBResponseError)

    def test_validation_exception_translation(self):
        from ddbmock.router.boto import _ddbmock_exception_to_boto_exception
        from boto.dynamodb.exceptions import DynamoDBValidationError
        from ddbmock.errors import ValidationException

        self.assertIsInstance(_ddbmock_exception_to_boto_exception(
                              ValidationException('taratata')),
                              DynamoDBValidationError)

    def test_conditions_exception_translation(self):
        from ddbmock.router.boto import _ddbmock_exception_to_boto_exception
        from boto.dynamodb.exceptions import DynamoDBConditionalCheckFailedError
        from ddbmock.errors import ConditionalCheckFailedException

        self.assertIsInstance(_ddbmock_exception_to_boto_exception(
                              ConditionalCheckFailedException('taratata')),
                              DynamoDBConditionalCheckFailedError)



# -*- coding: utf-8 -*-

import unittest

KEY_NAME = "ze chi name"
KEY_TYPE = "humanoid"

KEY_DEF = {
    'AttributeName': KEY_NAME,
    'AttributeType': KEY_TYPE,
}

KEY_VALUE = "TadaList"

KEY1 = {KEY_TYPE: KEY_VALUE}
KEY2 = {"Hey! me incompatible type: alien bacteria": KEY_VALUE}

class TestKey(unittest.TestCase):
    # actions
    def test_key_init(self):
        from ddbmock.database.key import Key

        k = Key(KEY_NAME, KEY_TYPE)
        self.assertEqual(KEY_NAME, k.name)
        self.assertEqual(KEY_TYPE, k.typename)

    def test_key_from_dict(self):
        from ddbmock.database.key import Key

        k = Key.from_dict(KEY_DEF)
        self.assertEqual(KEY_NAME, k.name)
        self.assertEqual(KEY_TYPE, k.typename)

    def test_key_read(self):
        from ddbmock.database.key import Key
        from ddbmock.errors import ValidationException

        k = Key(KEY_NAME, KEY_TYPE)
        self.assertEqual(KEY_VALUE, k.read(KEY1))
        self.assertRaises(ValidationException,
                          k.read,
                          KEY2)

    def test_to_dict(self):
        from ddbmock.database.key import Key

        k = Key(KEY_NAME, KEY_TYPE)

        expected = {
            "AttributeName": KEY_NAME,
            "AttributeType": KEY_TYPE,
        }

        self.assertEqual(expected, k.to_dict())



# -*- coding: utf-8 -*-

from .stat import Stat
from ddbmock import config
from threading import Timer
import logging

tp_stat = {
    'read': {},
    'write': {},
}

req_logger = logging.getLogger("Request logger")
req_logger.addHandler(logging.NullHandler())

tp_logger = logging.getLogger("Throughput logger")
tp_logger.addHandler(logging.NullHandler())

def push_read_throughput(table_name, value):
    if table_name not in tp_stat['read']:
        tp_stat['read'][table_name] = Stat("`%s` read throughput" % table_name,
                                               config.STAT_TP_SAMPLE,
                                               config.STAT_TP_AGGREG,
                                               tp_logger)
    tp_stat['read'][table_name].push(value)

def push_write_throughput(table_name, value):
    if table_name not in tp_stat['write']:
        tp_stat['write'][table_name] = Stat("`%s` write throughput" % table_name,
                                                config.STAT_TP_SAMPLE,
                                                config.STAT_TP_AGGREG,
                                                tp_logger)
    tp_stat['write'][table_name].push(value)

def schedule_action(delay, callback, args=[], kwargs={}):
    """Unless delays are explicitely disabled, start ``callback`` once ``delay``
    has expired. Otherwise, call it immediately.
    """
    if config.ENABLE_DELAYS:
        Timer(delay, callback, args, kwargs).start()
    else:
        callback(*args, **kwargs)



# -*- coding: utf-8 -*-

import unittest, json

TABLE_NAME1 = 'Table-HR'
TABLE_NAME2 = 'Table-H'
TABLE_NAME_404 = 'Waldo'
TABLE_RT = 45
TABLE_WT = 123
TABLE_RT2 = 10
TABLE_WT2 = 10
TABLE_HK_NAME = u'hash_key'
TABLE_HK_TYPE = u'N'
TABLE_RK_NAME = u'range_key'
TABLE_RK_TYPE = u'S'

HK_VALUE1 = u'123'
HK_VALUE2 = u'456'
HK_VALUE3 = u'789'
RK_VALUE1 = u'Waldo-1'
RK_VALUE2 = u'Waldo-2'
RK_VALUE3 = u'Waldo-3'
RK_VALUE4 = u'Waldo-4'
RK_VALUE5 = u'Waldo-5'


ITEM1 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE1},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE1},
    u'relevant_data': {u'S': u'tata'},
}
ITEM2 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE1},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE2},
    u'relevant_data': {u'S': u'tete'},
}
ITEM3 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE2},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE3},
    u'relevant_data': {u'S': u'titi'},
}
ITEM4 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE2},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE4},
    u'relevant_data': {u'S': u'toto'},
}
ITEM5 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE3},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE5},
    u'relevant_data': {u'S': u'tutu'},
}

HEADERS = {
    'x-amz-target': 'dynamodb_20111205.BatchWriteItem',
    'content-type': 'application/x-amz-json-1.0',
}

# Goal here is not to test the full API, this is done by the Boto tests
class TestBatchWriteItem(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.db import dynamodb
        from ddbmock.database.table import Table
        from ddbmock.database.key import PrimaryKey

        from ddbmock import main
        app = main({})
        from webtest import TestApp
        self.app = TestApp(app)

        dynamodb.hard_reset()

        hash_key = PrimaryKey(TABLE_HK_NAME, TABLE_HK_TYPE)
        range_key = PrimaryKey(TABLE_RK_NAME, TABLE_RK_TYPE)

        self.t1 = Table(TABLE_NAME1, TABLE_RT, TABLE_WT, hash_key, range_key)
        self.t2 = Table(TABLE_NAME2, TABLE_RT, TABLE_WT, hash_key, None)

        dynamodb.data[TABLE_NAME1]  = self.t1
        dynamodb.data[TABLE_NAME2]  = self.t2

        self.t1.put(ITEM1, {})
        self.t2.put(ITEM4, {})

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        dynamodb.hard_reset()

    def test_batch_write_item(self):
        from ddbmock.database.db import dynamodb

        request = {
            u"RequestItems": {
                TABLE_NAME1: [
                    {
                        u"DeleteRequest": {
                            u"Key": {
                                u"HashKeyElement": {TABLE_HK_TYPE: HK_VALUE1},
                                u"RangeKeyElement": {TABLE_RK_TYPE: RK_VALUE1},
                            },
                        },
                    },
                    {u"PutRequest": {u"Item": ITEM2}},
                    {u"PutRequest": {u"Item": ITEM3}},
                ],
                TABLE_NAME2: [
                    {
                        u"DeleteRequest": {
                            u"Key": {
                                u"HashKeyElement": {TABLE_HK_TYPE: HK_VALUE2},
                            },
                        },
                    },
                    {u"PutRequest": {u"Item":ITEM5}},
                ],
            }
        }

        expected = {
            "Responses": {
                TABLE_NAME1: {
                    "ConsumedCapacityUnits": 3
                },
                TABLE_NAME2: {
                    "ConsumedCapacityUnits": 2
                }
            }
        }

        # Protocol check
        res = self.app.post_json('/', request, HEADERS, status=200)
        self.assertEqual(expected, json.loads(res.body))
        self.assertEqual('application/x-amz-json-1.0; charset=UTF-8', res.headers['Content-Type'])



# -*- coding: utf-8 -*-

import unittest
import boto

TABLE_NAME = 'Table-HR'
TABLE_NAME2 = 'Table-H'
TABLE_NAME3 = 'Table-HR-size'
TABLE_NAME_404 = 'Waldo'
TABLE_RT = 45
TABLE_WT = 123
TABLE_RT2 = 10
TABLE_WT2 = 10
TABLE_HK_NAME = u'hash_key'
TABLE_HK_TYPE = u'N'
TABLE_RK_NAME = u'range_key'
TABLE_RK_TYPE = u'S'

HK_VALUE = u'123'
RK_VALUE = u'Decode this data if you are a coder'


ITEM = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE},
    u'relevant_data': {u'B': u'THVkaWEgaXMgdGhlIGJlc3QgY29tcGFueSBldmVyIQ=='},
}
ITEM2 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    TABLE_RK_NAME: {TABLE_RK_TYPE: RK_VALUE},
    u'irelevant_data': {u'B': u'WW91IHdpc2ggeW91IGNvdWxkIGNoYW5nZSB5b3VyIGpvYi4uLg=='},
}
ITEM3 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    u'relevant_data': {u'B': u'THVkaWEgaXMgdGhlIGJlc3QgY29tcGFueSBldmVyIQ=='},
}
ITEM3_EMPTY_FIELD = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    u'relevant_data': {u'B': u''},
}
ITEM3_EMPTY_SET = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    u'relevant_data': {u'B': u'THVkaWEgaXMgdGhlIGJlc3QgY29tcGFueSBldmVyIQ=='},
    u'empty_set': {u'NS': []},
}
ITEM4 = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    u'irelevant_data': {u'B': u'WW91IHdpc2ggeW91IGNvdWxkIGNoYW5nZSB5b3VyIGpvYi4uLg=='},
}
ITEM5 = {
    u'relevant_data': {u'B': u'THVkaWEgaXMgdGhlIGJlc3QgY29tcGFueSBldmVyIQ=='},
}


ITEM_BIG = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    u'relevant_data': {u'S': u'a'*1024},
}

ITEM_HUGE = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    u'relevant_data': {u'S': u'a'*64*1024},  # thsi field itself is.... too big
}

ITEM_OVER_H = {
    TABLE_HK_NAME: {u'S': 'a'*2049},
    TABLE_RK_NAME: {u'S': 'a'},
}
ITEM_MAX_H = {
    TABLE_HK_NAME: {u'S': 'a'*2048},
    TABLE_RK_NAME: {u'S': 'a'},
}
ITEM_OVER_R = {
    TABLE_HK_NAME: {u'S': 'a'},
    TABLE_RK_NAME: {u'S': 'a'*1025},
}
ITEM_MAX_R = {
    TABLE_HK_NAME: {u'S': 'a'},
    TABLE_RK_NAME: {u'S': 'a'*1024},
}

ITEM_REGRESION = {
    TABLE_HK_NAME: {TABLE_HK_TYPE: HK_VALUE},
    "Views": {"N": "0"},  # the one failing, sighs
    "forum_name": {"S": "Amazon dynamodb"},
    "Tags": {"SS": ["index", "primarykey", "table"]},
    "LastPostDateTime": {"S": "12/9/2011 11:36:03 PM"},
    "LastPostedBy": {"S": "User A"},
    "Answered": {"N": "0"},
    "Replies": {"N": "0"},
    "Message": {"S": "dynamodb thread 1 message text"},
    "Public": {"N": "1"},
    "subject": {"S": "dynamodb Thread 1"}
}

class TestPutItem(unittest.TestCase):
    def setUp(self):
        from ddbmock.database.db import dynamodb
        from ddbmock.database.table import Table
        from ddbmock.database.key import PrimaryKey

        dynamodb.hard_reset()

        hash_key = PrimaryKey(TABLE_HK_NAME, TABLE_HK_TYPE)
        hash_key2 = PrimaryKey(TABLE_HK_NAME, TABLE_RK_TYPE)
        range_key = PrimaryKey(TABLE_RK_NAME, TABLE_RK_TYPE)

        self.t1 = Table(TABLE_NAME, TABLE_RT, TABLE_WT, hash_key, range_key)
        self.t2 = Table(TABLE_NAME2, TABLE_RT, TABLE_WT, hash_key, None)
        self.t3 = Table(TABLE_NAME3, TABLE_RT, TABLE_WT, hash_key2, range_key)

        dynamodb.data[TABLE_NAME]  = self.t1
        dynamodb.data[TABLE_NAME2] = self.t2
        dynamodb.data[TABLE_NAME3] = self.t3

    def tearDown(self):
        from ddbmock.database.db import dynamodb
        from ddbmock import clean_boto_patch
        dynamodb.hard_reset()
        clean_boto_patch()

    def test_put_hr(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        db = connect_boto_patch()

        self.assertEqual({
                u'ConsumedCapacityUnits': 1,
            },
            db.layer1.put_item(TABLE_NAME, ITEM),
        )
        self.assertEqual(ITEM, self.t1.store[HK_VALUE, RK_VALUE])

        self.assertEqual({
                u'ConsumedCapacityUnits': 1,
                u'Attributes': ITEM,
            },
            db.layer1.put_item(TABLE_NAME, ITEM2, return_values=u'ALL_OLD'),
        )

    def test_put_h(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        db = connect_boto_patch()

        self.assertEqual({
                u'ConsumedCapacityUnits': 1,
            },
            db.layer1.put_item(TABLE_NAME2, ITEM3),
        )
        self.assertEqual(ITEM3, self.t2.store[HK_VALUE, False])

        self.assertEqual({
                u'ConsumedCapacityUnits': 1,
                u'Attributes': ITEM3,
            },
            db.layer1.put_item(TABLE_NAME2, ITEM4, return_values=u'ALL_OLD'),
        )

    def test_put_check_throughput_max_old_new(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        db = connect_boto_patch()

        self.assertEqual(
            {u'ConsumedCapacityUnits': 1},
            db.layer1.put_item(TABLE_NAME2, ITEM3),
        )
        self.assertEqual(
            {u'ConsumedCapacityUnits': 2},
            db.layer1.put_item(TABLE_NAME2, ITEM_BIG),
        )
        self.assertEqual(
            {u'ConsumedCapacityUnits': 2},
            db.layer1.put_item(TABLE_NAME2, ITEM3),
        )

    def test_put_h_empty_field_fail(self):
        # From http://docs.amazonwebservices.com/amazondynamodb/latest/developerguide/API_PutItem.html
        # Attribute values may not be null; string and binary type attributes must have lengths greater than zero; and set type attributes must not be empty. Requests with empty values will be rejected with a ValidationException.
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.dynamodb.exceptions import DynamoDBValidationError

        db = connect_boto_patch()

        self.assertRaises(DynamoDBValidationError,
                          db.layer1.put_item,
                          TABLE_NAME2, ITEM3_EMPTY_SET)

        self.assertNotIn((HK_VALUE, False), self.t2.store)

        self.assertRaises(DynamoDBValidationError,
                          db.layer1.put_item,
                          TABLE_NAME2, ITEM3_EMPTY_FIELD)

        self.assertNotIn((HK_VALUE, False), self.t2.store)

    def test_put_hr_404(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.exception import DynamoDBResponseError

        db = connect_boto_patch()

        self.assertRaisesRegexp(DynamoDBResponseError, 'ResourceNotFoundException',
                                db.layer1.put_item,
                                TABLE_NAME_404, ITEM)

    def test_put_h_404(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.exception import DynamoDBResponseError

        db = connect_boto_patch()

        self.assertRaisesRegexp(DynamoDBResponseError, 'ResourceNotFoundException',
                                db.layer1.put_item,
                                TABLE_NAME_404, ITEM3)

    def test_put_hr_missing_r(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.dynamodb.exceptions import DynamoDBValidationError

        db = connect_boto_patch()

        self.assertRaises(DynamoDBValidationError,
                          db.layer1.put_item,
                          TABLE_NAME, ITEM3)

    def test_put_h_missing_h(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.dynamodb.exceptions import DynamoDBValidationError

        db = connect_boto_patch()

        self.assertRaises(DynamoDBValidationError,
                          db.layer1.put_item,
                          TABLE_NAME2, ITEM5)

    def test_put_h_expect_no_exist(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.exception import DynamoDBResponseError

        db = connect_boto_patch()

        ddb_expected = {
            TABLE_HK_NAME: {u'Exists': False}
        }

        db.layer1.put_item(TABLE_NAME2, ITEM3, expected=ddb_expected)

        self.assertRaisesRegexp(DynamoDBResponseError, 'ConditionalCheckFailedException',
            db.layer1.put_item,
            TABLE_NAME2, ITEM4, expected=ddb_expected
        )
        self.assertEqual(ITEM3, self.t2.store[HK_VALUE, False])

    def test_put_h_expect_field_value(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.exception import DynamoDBResponseError

        db = connect_boto_patch()

        ddb_expected = {
            u'relevant_data': {
                u'Exists': True,
                u'Value': {u'B': u'THVkaWEgaXMgdGhlIGJlc3QgY29tcGFueSBldmVyIQ=='}
            }
        }

        db.layer1.put_item(TABLE_NAME2, ITEM3)
        self.assertEqual(ITEM3, self.t2.store[HK_VALUE, False])
        db.layer1.put_item(TABLE_NAME2, ITEM4, expected=ddb_expected)
        self.assertEqual(ITEM4, self.t2.store[HK_VALUE, False])
        self.assertRaisesRegexp(DynamoDBResponseError, 'ConditionalCheckFailedException',
            db.layer1.put_item,
            TABLE_NAME2, ITEM4, expected=ddb_expected
        )

    def test_put_oversized_h(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.dynamodb.exceptions import DynamoDBValidationError

        db = connect_boto_patch()

        db.layer1.put_item(TABLE_NAME3, ITEM_MAX_H)
        self.assertRaisesRegexp(DynamoDBValidationError, 'bytes',
            db.layer1.put_item,
            TABLE_NAME3, ITEM_OVER_H)

    def test_put_oversized_r(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.dynamodb.exceptions import DynamoDBValidationError

        db = connect_boto_patch()

        db.layer1.put_item(TABLE_NAME3, ITEM_MAX_R)
        self.assertRaisesRegexp(DynamoDBValidationError, 'bytes',
            db.layer1.put_item,
            TABLE_NAME3, ITEM_OVER_R)

    def test_put_oversized_item(self):
        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb
        from boto.dynamodb.exceptions import DynamoDBValidationError

        db = connect_boto_patch()

        self.assertRaisesRegexp(DynamoDBValidationError, 'Items.*smaller',
            db.layer1.put_item,
            TABLE_NAME2, ITEM_HUGE)

        self.assertNotIn((HK_VALUE, False), self.t2.store)

    def test_put_boto_intergration(self):
        # This item comes directly from boto intergration test suite
        # we do no assertion. It should "just work"

        from ddbmock import connect_boto_patch
        from ddbmock.database.db import dynamodb

        db = connect_boto_patch()
        db.layer1.put_item(TABLE_NAME2, ITEM_REGRESION)



# -*- coding: utf-8 -*-

from .types import table_name, item_schema, get_key_schema, Any

post = {
    u"RequestItems": {
        table_name: [Any(
            {
                u"PutRequest": {
                    u"Item": item_schema,
                }
            },{
                u"DeleteRequest": {
                    u"Key": get_key_schema,
                }
            },
        )],
    },
}
